# pylint: disable=no-member
"""Module of drift types."""
from dataclasses import dataclass
from io import StringIO
from typing import Dict, List, Optional

import pandas as pd
from pydantic import BaseModel
from pygments import formatters, highlight, lexers


class FeatureSummary(BaseModel):
    """Provides a summary of a features dataframe."""

    #: int: Number of samples (rows).
    n_rows: int
    #: int: Number of features (columns).
    n_features: int


class LabelSummary(BaseModel):
    """Provides a summary of a labels series."""

    #: int: Number of samples (rows).
    n_rows: int
    #: int: Number of distinct labels. For example, for binary data, this would
    #: be equal to 2.
    n_labels: int


class LatentSummary(BaseModel):
    """Provides a summary of a latents dataframe."""

    #: int: Number of samples (rows).
    n_rows: int
    #: int: Number of latent features (columns).
    n_latents: int


class ShapeSummary(BaseModel):
    """Provides a summary of the object shapes in a dataset of features,
    labels and latents.
    """

    #: FeatureSummary: Features shape summary.
    features: FeatureSummary
    #: LabelSummary: Labels shape summary.
    labels: LabelSummary
    #: Optional[LatentSummary]: Optional latents shape summary.
    latents: Optional[LatentSummary]


class BaselineSummary(BaseModel):
    """Class for storing a shape summary with JSON string representation."""

    #: ShapeSummary: A shape summary instance of a dataset.
    shapes: ShapeSummary

    def __str__(self) -> str:
        """Generates a JSON string for shape summary of dataset."""
        output = self.json(indent=2)
        return str(
            highlight(output, lexers.JsonLexer(), formatters.TerminalFormatter())
        )


@dataclass
class Dataset:
    """Class for representing a drift dataset."""

    #: pd.DataFrame: A combined dataframe of input features and ground truth labels.
    features: pd.DataFrame
    #: pd.Series: A series of predicted labels from a model.
    labels: pd.Series
    #: Optional[pd.DataFrame]: An optional dataframe of latent variables per sample.
    latents: Optional[pd.DataFrame] = None

    @property
    def feature_names(self) -> List[str]:
        """Returns a list of features dataframe columns.

        Returns:
            List[str]: A list of feature column names as strings.

        """
        return list(self.features.columns)

    def unify(self) -> pd.DataFrame:
        """Returns a column-wise concatenated dataframe of features, labels and
        latents.

        Returns:
            pd.DataFrame: Column-wise concatenated dataframe of features,
                labels and latents.

        """
        return pd.concat(
            [
                self.features,
                pd.DataFrame(self.labels),
                self.latents,
            ],
            axis=1,
        )


@dataclass
class StructuredResult:
    """A type for representing a result from the hypothesis tests module."""

    #: str: Name of the scoring method used.
    method_name: str
    #: Dict[str, Dict[str, float]]: Dictionary of results with keys as
    #: `feature_name` or, if for a unified dataset, "single_value".
    #: Values are a dictionary containing the result statistic and p-value (if
    #: available) for a given `method_name`.
    results: Dict[str, Dict[str, float]]

    def __repr__(self) -> str:
        output = StringIO()
        print(f"Method: {self.method_name}", file=output)
        for result, result_dict in self.results.items():
            print(f"  {result}", file=output)
            for (result_key, result_value) in result_dict.items():
                print(f"{result_key: >15}: {result_value:>10.2e}", file=output)
            print(file=output)
        return output.getvalue()



"""Module with class to filter a dataset."""
from enum import Enum, auto
from typing import Any, List, Optional

import numpy as np
import pandas as pd

from learning_machines_drift.types import Dataset


class Comparison(Enum):
    """Comparison enum for 'LESS', 'GREATER' and 'EQUAL' cases."""

    LESS = auto()
    GREATER = auto()
    EQUAL = auto()


class Condition:  # pylint: disable=too-few-public-methods
    """Condition class comprising of a 'comparison' and a 'value'."""

    comparison: Comparison
    value: Any

    def __init__(self, comparison_str: str, value: Any):
        """Init condition from passed string and value.

        Args:
            comparison_str (str): One of 'equal', 'less', 'greater'.
            value (Any): The value for comparison.

        """
        if comparison_str == "less":
            self.comparison = Comparison.LESS
        elif comparison_str == "greater":
            self.comparison = Comparison.GREATER
        elif comparison_str == "equal":
            self.comparison = Comparison.EQUAL
        else:
            raise ValueError(
                f"'{comparison_str}' is not implemented. "
                "Please choose one of 'equal', 'less', 'greater'."
            )
        self.value = value


class Filter:  # pylint: disable=too-few-public-methods
    """Filter class.

    Filters a given dataset through an `AND` operation applied across all
    passed conditions.
    """

    #: dict[str, List[Condition]]: Dict with key (variable) and value as a
    #: list of (condition, value) to be used for filtering.
    conditions: Optional[dict[str, List[Condition]]]

    def __init__(self, conditions: Optional[dict[str, List[Condition]]]):
        """Initialize a dict with variable keys and (condition, value) values.

        Args:
            conditions (dict[str, List[Condition]]): Dict with key (variable)
                and value as a list of (condition, value) to be used for
                filtering.

        """
        self.conditions = conditions

    def _filter_df(
        self, df: pd.DataFrame, variable: str, condition: Condition
    ) -> pd.DataFrame:
        """Subset a dataframe given condition and value."""
        if condition.comparison == Comparison.LESS:
            return df.loc[df[variable] < condition.value]
        if condition.comparison == Comparison.GREATER:
            return df.loc[df[variable] > condition.value]
        if condition.comparison == Comparison.EQUAL:
            return df.loc[df[variable] == condition.value]
        raise ValueError(
            f"'{condition.comparison}' is not implemented. "
            "Please choose one of 'equal', 'less', 'greater'."
        )

    def _filter_series(
        self,
        series: pd.Series,
        condition: Condition,
    ) -> pd.Series:
        """Subset a series given condition and value."""
        if condition.comparison == Comparison.LESS:
            return series[series < condition.value]
        if condition.comparison == Comparison.GREATER:
            return series[series > condition.value]
        if condition.comparison == Comparison.EQUAL:
            return series[series == condition.value]
        raise ValueError(
            f"'{condition.comparison}' is not implemented. "
            "Please choose one of 'equal', 'less', 'greater'."
        )

    def transform(self, dataset: Dataset) -> Dataset:
        """Transform the passed dataset given filter.

        Args:
            dataset (Dataset): the dataset to be filtered.

        Returns:
            Dataset: transformed dataset given filters.
        """
        if self.conditions is None:
            return dataset

        # Filter the dataframe or series of the variables
        for (variable, conditions) in self.conditions.items():
            for condition in conditions:
                if variable in dataset.features.columns:
                    dataset.features = self._filter_df(
                        dataset.features, variable, condition
                    )
                elif variable == dataset.labels.name:
                    dataset.labels = self._filter_series(dataset.labels, condition)
                elif dataset.latents is not None:
                    if variable in dataset.latents.columns:
                        dataset.latents = self._filter_df(
                            dataset.latents, variable, condition
                        )

        # Select only the common idx
        common_idx = np.intersect1d(
            dataset.features.index.to_numpy(),
            dataset.labels.index.to_numpy(),
        )
        if dataset.latents is not None:
            common_idx = np.intersect1d(
                common_idx,
                dataset.latents.index.to_numpy(),
            )

        # Update dataset with common idx
        dataset.features = dataset.features.loc[common_idx]
        dataset.labels = dataset.labels.loc[common_idx]
        if dataset.latents is not None:
            dataset.latents = dataset.latents.loc[common_idx]

        # Return filtered dataset
        return dataset



"""Tests for drift detector."""
# pylint: disable=W0621,too-many-locals

import pathlib
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from pytest_mock import MockerFixture

from learning_machines_drift import Monitor, Registry
from learning_machines_drift.backends import FileBackend
from learning_machines_drift.datasets import example_dataset
from learning_machines_drift.display import Display
from learning_machines_drift.exceptions import ReferenceDatasetMissing
from learning_machines_drift.filter import Comparison, Condition, Filter
from learning_machines_drift.types import Dataset, StructuredResult

N_FEATURES = 3
N_LABELS = 2
N_LATENTS = 1

# Range of rows for most tests
N_ROWS = [5, 10, 100, 1000]


@pytest.fixture()
def registry(mocker: MockerFixture) -> Registry:
    """Returns a registry that writes data to a temporary directory."""
    det = Registry(tag="test", expect_latent=True)
    mocker.patch.object(det, "backend")
    return det


@pytest.fixture()
def monitor(mocker: MockerFixture) -> Monitor:
    """Returns a registry that writes data to a temporary directory."""
    meas = Monitor(tag="test")
    mocker.patch.object(meas, "backend")
    return meas


@pytest.fixture()
def registry_with_ref_data(registry: Registry) -> Callable[[int], Registry]:
    """Returns a registry with a reference dataset registered which
    writes data to a temporary directory.
    """

    def _registry_with_ref_data(n_rows: int) -> Registry:
        """Returns registry with saved reference data."""
        features_df, labels_df, latents_df = example_dataset(n_rows)
        # When we register the dataset
        registry.register_ref_dataset(
            features=features_df, labels=labels_df, latents=latents_df
        )
        return registry

    return _registry_with_ref_data


def registry_with_log_data(registry: Registry, num_rows: int) -> Registry:
    """Returns a registry with a logged datasets registered which
    writes data to a temporary directory.
    """
    num_iterations = 1
    for _ in range(num_iterations):
        (
            new_features_df,
            new_predictions_series,
            new_latents_df,
        ) = example_dataset(num_rows)
        with registry:
            # And we have logged features, labels and latent
            registry.log_features(new_features_df)
            registry.log_labels(new_predictions_series)
            registry.log_latents(new_latents_df)
    return registry


def registry_with_all_data(
    path: pathlib.Path,
    n_rows: int,
    n_iterations: int = 1,
    to_drop: Optional[List[str]] = None,
) -> Registry:
    """Generates a registry with registered and reference data."""
    features_df, labels_df, latents_df = example_dataset(n_rows)
    det = Registry(tag="test", backend=FileBackend(path))
    det.register_ref_dataset(features=features_df, labels=labels_df, latents=latents_df)
    for _ in range(n_iterations):
        new_features_df, new_predictions_series, new_latents_df = example_dataset(
            n_rows
        )
        if to_drop is not None:
            new_features_df = new_features_df.drop(columns=to_drop)
        with det:
            det.log_features(new_features_df)
            det.log_labels(new_predictions_series)
            det.log_latents(new_latents_df)
    return det


@pytest.mark.parametrize("n_rows", N_ROWS)
def test_reference_dataset(
    registry_with_ref_data: Callable[[int], Registry], n_rows: int
) -> None:
    """Tests whether registry has expected reference data."""

    # Given we have a reference dataset
    det: Registry = registry_with_ref_data(n_rows)

    # When we get a summary of the reference set
    summary = det.ref_summary()

    # Then we can access summary information
    assert summary.shapes.features.n_rows == n_rows
    assert summary.shapes.features.n_features == N_FEATURES
    assert summary.shapes.labels.n_rows == n_rows
    assert summary.shapes.labels.n_labels == N_LABELS
    if summary.shapes.latents is not None:
        assert summary.shapes.latents.n_rows == n_rows
        assert summary.shapes.latents.n_latents == N_LATENTS

    # And we saved the data to the backend
    det.backend.save_reference_dataset.assert_called_once()  # type: ignore


def test_ref_summary_no_dataset(registry) -> None:  # type: ignore
    """Tests that the correct exception is raised upon registry with no
    reference dataset."""

    # Given a registry with no reference dataset registered

    # When we get the reference dataset summary
    # Then raise an exception
    with pytest.raises(ReferenceDatasetMissing):
        _ = registry.ref_summary()

    # And we should not have saved any reference data
    registry.backend.save_reference_dataset.assert_not_called()


@pytest.mark.parametrize("n_rows", N_ROWS)
def test_all_registered(
    registry_with_ref_data: Callable[[int], Registry], n_rows: int
) -> None:
    """Tests whether all expected data is registered."""
    # Given we have registered a reference dataset
    det = registry_with_ref_data(n_rows)
    det = registry_with_log_data(det, n_rows)

    # Then we can ensure that everything is registered
    assert det.all_registered()

    # And we saved a reference dataset
    det.backend.save_reference_dataset.assert_called_once()  # type: ignore

    # And we saved the logged features
    det.backend.save_logged_features.assert_called_once()  # type: ignore

    # And we saved the logged labels
    det.backend.save_logged_labels.assert_called_once()  # type: ignore


@pytest.mark.parametrize("n_rows", [10])
def test_statistics_summary(tmp_path: pathlib.Path, n_rows: int) -> None:
    """Tests whether metrics are applied to all columns of unifed
    dataset as expected."""
    det: Registry = registry_with_all_data(tmp_path, n_rows)
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()
    res = monitor.metrics.scipy_kolmogorov_smirnov()

    assert isinstance(res, StructuredResult)
    assert res.method_name == "scipy_kolmogorov_smirnov"
    assert list(res.results.keys()) == list(det.registered_dataset.unify().columns)


@pytest.mark.parametrize("n_rows", N_ROWS)
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_summary_statistic_list(tmp_path: pathlib.Path, n_rows: int) -> None:
    """Tests application metrics from a monitor."""
    det: Registry = registry_with_all_data(tmp_path, n_rows)
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()
    h_test_dispatcher: Dict[str, Any] = {
        "scipy_kolmogorov_smirnov": monitor.metrics.scipy_kolmogorov_smirnov,
        "scipy_mannwhitneyu": monitor.metrics.scipy_mannwhitneyu,
        "boundary_adherence": monitor.metrics.get_boundary_adherence,
        "range_coverage": monitor.metrics.get_range_coverage,
        "logistic_detection": monitor.metrics.logistic_detection,
    }
    for h_test_name, h_test_fn in h_test_dispatcher.items():
        res = h_test_fn()
        assert isinstance(res, StructuredResult)
        assert res.method_name == h_test_name
        if list(res.results.keys())[0] == "single_value":
            assert isinstance(res.results["single_value"], dict)
        else:
            assert list(res.results.keys()) == list(
                det.registered_dataset.unify().columns
            )


@pytest.mark.parametrize("n_rows", N_ROWS)
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_with_noncommon_columns(tmp_path: pathlib.Path, n_rows: int) -> None:
    """Tests the application of metrics to the intersection of columns
    when there are differing columns in the reference and registered dataset.
    """
    det: Registry = registry_with_all_data(tmp_path, n_rows, to_drop=["age"])
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()
    h_test_dispatcher: Dict[str, Any] = {
        "scipy_kolmogorov_smirnov": monitor.metrics.scipy_kolmogorov_smirnov,
        "scipy_mannwhitneyu": monitor.metrics.scipy_mannwhitneyu,
        "scipy_permutation": monitor.metrics.scipy_permutation,
        "logistic_detection": monitor.metrics.logistic_detection,
        "logistic_detection_f1": partial(
            monitor.metrics.logistic_detection, score_type="f1"
        ),
        "logistic_detection_roc_auc": partial(
            monitor.metrics.logistic_detection, score_type="roc_auc"
        ),
    }
    for h_test_name, h_test_fn in h_test_dispatcher.items():
        res = h_test_fn()
        assert isinstance(res, StructuredResult)
        assert res.method_name == h_test_name
        if list(res.results.keys())[0] == "single_value":
            assert isinstance(res.results["single_value"], dict)
        else:
            assert list(res.results.keys()) == list(
                det.registered_dataset.unify().columns
            )


@pytest.mark.parametrize("n_rows", [10])
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_display(tmp_path: pathlib.Path, n_rows: int) -> None:
    """Tests whether the display class returns the expected types."""
    registry_with_all_data(tmp_path, n_rows)
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()

    # Subset of h_tests sufficient for testing plotting
    h_test_dispatcher: Dict[str, Callable[..., StructuredResult]] = {
        "scipy_kolmogorov_smirnov": monitor.metrics.scipy_kolmogorov_smirnov,
        "scipy_mannwhitneyu": monitor.metrics.scipy_mannwhitneyu,
    }

    for _, h_test_fn in h_test_dispatcher.items():
        # Get result from scoring
        structured_result: StructuredResult = h_test_fn(verbose=False)

        # Check Display produces a dataframe
        df = Display().table(structured_result, verbose=False)
        assert isinstance(df, pd.DataFrame)
        # Check the correct shape
        assert df.shape == (len(structured_result.results), 2)

        # Check Display returns a plot and array of axes
        fig, axs = Display().plot(structured_result)
        assert isinstance(fig, plt.Figure)
        assert isinstance(axs, np.ndarray)
        for ax in axs.flatten():
            assert isinstance(ax, plt.Axes)


@pytest.mark.parametrize("n_rows", N_ROWS)
def test_load_all_logged_data(tmp_path: pathlib.Path, n_rows: int) -> None:
    """Tests whether logging of data and that upon reload the data has the
    correct shape."""
    registry_with_all_data(tmp_path, n_rows)
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()
    assert monitor.registered_dataset is not None
    recovered_dataset: Dataset = monitor.registered_dataset
    dimensions: Tuple[int, int] = recovered_dataset.unify().shape
    assert dimensions[0] == n_rows
    assert dimensions[1] == (3 + 1 + 1)


def test_condition() -> None:
    """Test for creating condition instances."""
    condition = Condition("less", 5)
    assert condition.comparison == Comparison.LESS
    assert condition.value == 5

    with pytest.raises(Exception):
        condition = Condition("Equal", 5)

    with pytest.raises(Exception):
        condition = Condition("Greater", 5)


# Only test large number of samples as stochastic samples
@pytest.mark.parametrize("n_rows", [1000])
def test_load_data_filtered(tmp_path: pathlib.Path, n_rows: int) -> None:
    """Tests whether a filter applied to load data from registry correctly
    filters data."""
    registry_with_all_data(tmp_path, n_rows)
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()

    # Make a drift filter
    filter_dict: Dict[str, List[Condition]] = dict(
        [
            ("age", [Condition("less", 0.2)]),
            ("height", [Condition("greater", -0.1), Condition("less", 0.5)]),
            ("ground-truth-label", [Condition("equal", 0)]),
            ("latents", [Condition("greater", 0.6)]),
        ]
    )
    drift_filter = Filter(filter_dict)

    # Load data (unfiltered datasets)
    monitor_unfiltered = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()

    # Load data (filtered datasets)
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data(drift_filter)

    # Assert conditions on unfiltered and filtered datasets
    for (ref_dataset, reg_dataset, assertion_bool) in [
        (monitor_unfiltered.ref_dataset, monitor_unfiltered.registered_dataset, False),
        (monitor.ref_dataset, monitor.registered_dataset, True),
    ]:
        assert ref_dataset is not None and reg_dataset is not None
        assert ref_dataset.latents is not None and reg_dataset.latents is not None
        assert ref_dataset.features.shape[0] == ref_dataset.labels.shape[0]
        assert ref_dataset.labels.shape[0] == ref_dataset.latents.shape[0]
        assert reg_dataset.features.shape[0] == reg_dataset.labels.shape[0]
        assert reg_dataset.labels.shape[0] == reg_dataset.latents.shape[0]
        assert ref_dataset.features["age"].lt(0.2).all() == assertion_bool
        assert reg_dataset.features["age"].lt(0.2).all() == assertion_bool
        assert ref_dataset.features["height"].gt(-0.1).all() == assertion_bool
        assert reg_dataset.features["height"].gt(-0.1).all() == assertion_bool
        assert ref_dataset.features["height"].lt(0.5).all() == assertion_bool
        assert reg_dataset.features["height"].lt(0.5).all() == assertion_bool
        assert ref_dataset.labels.eq(0).all() == assertion_bool
        assert reg_dataset.labels.eq(0).all() == assertion_bool
        assert ref_dataset.latents["latents"].gt(0.6).all() == assertion_bool
        assert reg_dataset.latents["latents"].gt(0.6).all() == assertion_bool


@pytest.mark.parametrize("n_rows", [10])
def test_dataset_types(tmp_path: pathlib.Path, n_rows: int) -> None:
    """Tests whether the reference dataset components have the expected types."""
    registry_with_all_data(tmp_path, n_rows)
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()

    assert monitor.ref_dataset is not None
    assert isinstance(monitor.ref_dataset.features, pd.DataFrame)
    assert isinstance(monitor.ref_dataset.labels, pd.Series)
    assert isinstance(monitor.ref_dataset.latents, pd.DataFrame)


@pytest.mark.parametrize("n_rows", N_ROWS)
def test_sdmetrics(tmp_path: pathlib.Path, n_rows: int) -> None:
    """Tests SD metrics."""
    registry_with_all_data(tmp_path, n_rows)
    monitor = Monitor(tag="test", backend=FileBackend(tmp_path)).load_data()

    for (sd_metric_name, sd_metric_fn) in [
        ("boundary_adherence", monitor.metrics.get_boundary_adherence),
        ("range_coverage", monitor.metrics.get_range_coverage),
    ]:
        result: StructuredResult = sd_metric_fn()
        assert result.method_name == sd_metric_name
        assert isinstance(result.results, dict)
        assert len(result.results.keys()) == 5
        assert next(iter(result.results)) == "age"
        result_age = result.results["age"]
        assert isinstance(result_age, dict)
        assert next(iter(result_age.keys())) == "statistic"
        assert result_age["statistic"] >= 0.0 and result_age["statistic"] <= 1.0



"""Class for scoring drift between reference and registered datasets."""

import textwrap
from enum import Enum
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import stats
from sdmetrics.single_column import BoundaryAdherence, RangeCoverage
from sdmetrics.single_table import LogisticDetection
from sdmetrics.utils import HyperTransformer
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import f1_score, roc_auc_score
from sklearn.model_selection import StratifiedKFold

from learning_machines_drift.types import Dataset, StructuredResult


class Wrapper(Enum):
    """Enum for specifying the calculation type."""

    TYPE_TUPLE = 1
    TYPE_OTHER = 2
    TYPE_SDMETRIC = 3


class Metrics:
    """A class with metrics for scoring data drift between registered and
    reference datasets.

    Attributes:
        reference_dataset (Dataset): Reference datastet for drift measures.
        registered_dataset (Dataset): Registered/logged datastet for drift
            measures.
        random_state (Optional[int]): Optional seeding for reproducibility.
    """

    def __init__(
        self,
        reference_dataset: Dataset,
        registered_dataset: Dataset,
        random_state: Optional[int] = None,
    ) -> None:
        """Initialize with registered and reference and optional seed.

        Args:
            reference_dataset (Dataset): Reference datastet for drift measures.
            registered_dataset (Dataset): Registered/logged datastet for drift
                measures.
            random_state (Optional[int]): Optional seeding for reproducibility.

        """
        self.reference_dataset = reference_dataset
        self.registered_dataset = registered_dataset
        self.random_state = random_state

    @staticmethod
    def _format_about_str(method: str, description: str, line_len: int = 79) -> str:
        """Takes methods and description and returns a string to print.

        Args:
            method (str): Name of the method.
            description (str): Description of the method.

        Returns:
            str: Formatted string.
        """
        return "\n".join(
            [
                "",
                textwrap.fill(f"Method: {method}".replace("\n", " "), line_len),
                textwrap.fill(
                    f"Description: {description}".replace("\n", " "), line_len
                ),
                "",
            ]
        )

    @staticmethod
    def _to_dict(obj: object) -> Dict[str, float]:
        """Convert an object to a dict."""
        obj_as_dict = {}
        attributes = [
            attribute for attribute in dir(obj) if attribute in ["statistic", "pvalue"]
        ]
        for attribute in attributes:
            obj_as_dict[attribute] = getattr(obj, attribute)
        return obj_as_dict

    def _calc(
        self,
        func: Callable[..., Any],
        subset: Optional[List[str]] = None,
        wrapper: Wrapper = Wrapper.TYPE_OTHER,
    ) -> Any:
        """Method for calculating statistic and pvalue from a passed scoring
        function.

        Args:
            func (Callable[..., Any]): Function that takes two series and
                returns a dictionary or named tuple containing scores.
            subset (List[str]): List of feature names to subset scores over.

        Returns:
            results (dict): Dictionary with features as keys and scores as
                values.
        """

        def call_func(
            ref_col: pd.Series,
            reg_col: pd.Series,
            wrapper: Wrapper = Wrapper.TYPE_OTHER,
        ) -> Dict[str, float]:
            """Returns a dict of statistic and p-value (if available)."""
            if wrapper is Wrapper.TYPE_TUPLE:
                result = func((ref_col, reg_col))
            elif wrapper is Wrapper.TYPE_SDMETRIC:
                result = func(real_data=ref_col, synthetic_data=reg_col)
                result = {"statistic": result}
            else:
                result = func(ref_col, reg_col)

            if not isinstance(result, dict):
                result = self._to_dict(result)

            result_dict: Dict[str, float] = result
            return result_dict

        results: Dict[str, Dict[str, float]] = {}

        if subset is not None:
            # If subset, only loop over the subset
            columns_to_calc: List[str] = subset
        else:
            # Get all columns from features, labels, latents
            columns_to_calc = self.reference_dataset.unify().columns.to_list()

        for col_name in columns_to_calc:
            # Check if in features
            if self.reference_dataset is not None:
                if (col_name in self.reference_dataset.features.columns) and (
                    col_name in self.registered_dataset.features.columns
                ):
                    ref_col = self.reference_dataset.features[col_name]
                    reg_col = self.registered_dataset.features[col_name]
                # Check if labels
                elif (col_name == self.reference_dataset.labels.name) and (
                    col_name == self.registered_dataset.labels.name
                ):
                    ref_col = self.reference_dataset.labels
                    reg_col = self.registered_dataset.labels
                # Check if in latents
                elif (
                    (self.reference_dataset.latents is not None)
                    and (col_name in self.reference_dataset.latents.columns)
                    and (self.registered_dataset.latents is not None)
                    and (col_name in self.registered_dataset.latents.columns)
                ):
                    ref_col = self.reference_dataset.latents[col_name]
                    reg_col = self.registered_dataset.latents[col_name]
                else:
                    print(
                        "Error:",
                        ValueError(
                            f"'{col_name}' is not in features, labels or latents."
                        ),
                    )
                    continue
            else:
                raise ValueError("Reference dataset is None.")
            # Run calc and update dictionary
            results[col_name] = call_func(ref_col, reg_col, wrapper)

        return results

    def _get_unified_subsets(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Unify both datasets and return dataframes with common columns."""

        def get_intersect(data1: pd.DataFrame, data2: pd.DataFrame) -> List[str]:
            """Get list of common columns to both dataframes."""
            return list(set(data1.columns) & set(data2.columns))

        unified_ref = self.reference_dataset.unify()
        unified_reg = self.registered_dataset.unify()
        subset: List[str] = get_intersect(unified_ref, unified_reg)
        return (unified_ref[subset], unified_reg[subset])

    def scipy_kolmogorov_smirnov(self, verbose: bool = True) -> StructuredResult:
        """Calculates feature-wise two-sample Kolmogorov-Smirnov test for
        goodness of fit. Assumes continuous underlying distributions but
        scores are still interpretable if data is approximately continuous.

        Args:
            verbose (bool): Boolean for verbose output to stdout.

        Returns:
            results (dict): Dictionary of statistics and  p-values by feature.
        """
        method = "SciPy Kolmogorov Smirnov"
        description = ""
        about_str = self._format_about_str(method=method, description=description)

        results = self._calc(stats.ks_2samp)
        if verbose:
            print(about_str)

        result_dict: Dict[str, Dict[str, float]] = results
        structured_result = StructuredResult("scipy_kolmogorov_smirnov", result_dict)
        return structured_result

    def scipy_mannwhitneyu(self, verbose: bool = True) -> StructuredResult:
        """Calculates feature-wise Mann-Whitney U test, a nonparametric test of
        the null hypothesis that the distribution underlying sample x is the
        same as the distribution underlying sample y. Provides a test for the
        difference in location of two distributions. Assumes continuous
        underlying distributions but scores are still interpretable if data
        is approximately continuous.

        Args:
            verbose (bool): Boolean for verbose output to stdout.

        Returns:
            results (dict): Dictionary of statistics and  p-values by feature.

        """
        method = "SciPy Mann-Whitney U"
        description = (
            "Non-parameric test between independent samples comparing their location."
        )
        about_str = self._format_about_str(method=method, description=description)

        results = self._calc(stats.mannwhitneyu)
        if verbose:
            print(about_str)

        result_dict: Dict[str, Dict[str, float]] = results
        structured_result = StructuredResult("scipy_mannwhitneyu", result_dict)
        return structured_result

    def scipy_permutation(
        self,
        agg_func: Callable[..., float] = np.mean,
        verbose: bool = True,
    ) -> StructuredResult:
        """Performs feature-wise permutation test with default statistic to
        measure differences under permutations of labels as the mean.

        Args:
            func (Callable[..., float]): Function for comparing two samples.
            verbose (bool): Print outputs
        Returns:
            results (dict): Dictionary with keys as features and values as
            scipy.stats.permutation_test object with test results.
        """
        method = f"SciPy Permutation Test (test function: {agg_func.__name__})"
        description = (
            "Performs permutation test on all features with passed stat_fn "
            "measuring the difference between samples."
        )
        about_str = self._format_about_str(method=method, description=description)

        def statistic(
            lhs: npt.ArrayLike,
            rhs: npt.ArrayLike,
            axis: int = 0,
        ) -> float:
            """Statistic for evaluating the difference between permuted
            samples.
            """
            return agg_func(lhs, axis=axis) - agg_func(rhs, axis=axis)

        func = partial(
            stats.permutation_test,
            statistic=statistic,
            permutation_type="independent",
            alternative="two-sided",
            n_resamples=9999,
            random_state=self.random_state,
        )

        results = self._calc(func, wrapper=Wrapper.TYPE_TUPLE)

        if verbose:
            print(about_str)

        result_dict: Dict[str, Dict[str, float]] = results
        structured_result = StructuredResult("scipy_permutation", result_dict)
        return structured_result

    # pylint: disable=invalid-name
    def logistic_detection(  # pylint: disable=too-many-locals, too-many-branches
        self,
        normalize: bool = False,
        score_type: Optional[str] = None,
        seed: Optional[int] = None,
        verbose: bool = True,
    ) -> StructuredResult:
        """Calculates a measure of similarity using fitted logistic regression
        to predict reference or registered label. SD metrics package
        `source <https://github.com/sdv-dev/SDMetrics/blob/v0.6.0/sdmetrics/single_table/detection/base.py#L46-L91>`_ # pylint: disable=line-too-long
        is adapted to permit optional `score_type` and `seed` to be given allowing
        alternative and reproducible metrics.

        `score_type` can be:
            - None: defaults to scoring of `logistic_detection` method.
            - "f1": Cross-validated F1 score with  0.5 threshold.
            - "roc_auc": Cross-validated receiver operating characteristic (area under the curve).

        Args:
            score_type (Optional[str]): None for default or string; "f1" and
                "roc_auc" currently implemented.
            seed (Optional[int]): Optional integer for reproducibility of
                scoring as cross-validation performed.
            verbose (bool): Boolean for verbose output to stdout.

        Returns:
            results (float): Score providing an overall similarity measure of
                reference and registered datasets.
        """
        if score_type is None:
            method = f"Logistic Detection (scoring: standard, normalize: {normalize})"
        else:
            method = (
                f"Logistic Detection (scoring: {score_type}, normalize: {normalize})"
            )
        description = (
            "Detection metric based on a LogisticRegression classifier from "
            "scikit-learn with custom scoring."
        )
        about_str = self._format_about_str(method=method, description=description)

        if verbose:
            print(about_str)

        # Get unified subsets
        unified_ref_subset, unified_reg_subset = self._get_unified_subsets()

        # Transform data for fitting using SD metrics HyperTransformer
        ht = HyperTransformer()
        transformed_reference_data = ht.fit_transform(unified_ref_subset).to_numpy()
        transformed_registered_data = ht.transform(unified_reg_subset).to_numpy()

        X = np.concatenate([transformed_reference_data, transformed_registered_data])
        y = np.hstack(
            [
                np.ones(len(transformed_reference_data)),
                np.zeros(len(transformed_registered_data)),
            ]
        )
        if np.isin(X, [np.inf, -np.inf]).any():
            X[np.isin(X, [np.inf, -np.inf])] = np.nan

        try:
            scores = []
            kf = StratifiedKFold(n_splits=3, shuffle=True, random_state=seed)
            lr = LogisticRegression(solver="lbfgs")
            for train_index, test_index in kf.split(X, y):
                lr.fit(X[train_index], y[train_index])
                y_pred = lr.predict_proba(X[test_index])[:, 1]

                if score_type is None:
                    roc_auc = roc_auc_score(y[test_index], y_pred)
                    score = max(0.5, roc_auc) * 2 - 1
                elif score_type == "roc_auc":
                    score = roc_auc_score(y[test_index], y_pred)
                elif score_type == "f1":
                    score = f1_score(y[test_index], (y_pred > 0.5) * 1)
                else:
                    raise NotImplementedError(f"{score_type} not implemented.")
                scores.append(score)
        except ValueError as err:
            raise ValueError(
                f"DetectionMetric: Unable to be fit with error {err}"
            ) from err

        if score_type is None:
            # SDMetrics approach to scoring takes 1 - mean:
            # https://github.com/sdv-dev/SDMetrics/blob/master/sdmetrics/single_table/detection/base.py#L89
            results_key: str = "logistic_detection"
            results: float = 1 - np.mean(scores)
        else:
            # Custom metrics assume the mean of the scores
            results_key = f"logistic_detection_{score_type}"
            results = np.mean(scores)

        if normalize and score_type is None:
            results = LogisticDetection.normalize(results)

        result_dict: Dict[str, Dict[str, float]] = {
            "single_value": {"statistic": results}
        }
        structured_result = StructuredResult(results_key, result_dict)
        return structured_result

    # pylint: enable=invalid-name

    def get_boundary_adherence(
        self,
    ) -> StructuredResult:
        """For each feature the proportion of registered data that lies within
        the minimum and maximum of the reference dataset.

        See `SDMetrics <https://docs.sdv.dev/sdmetrics/metrics/metrics-glossary/boundaryadherence>`_
        for further details.

        Returns:
            StructuredResult: The boundary adherence of the registered dataset
                compared to the reference dataset.

        """
        results: Dict[str, Dict[str, float]] = self._calc(
            BoundaryAdherence.compute, wrapper=Wrapper.TYPE_SDMETRIC
        )
        structured_result = StructuredResult("boundary_adherence", results)
        return structured_result

    def get_range_coverage(self) -> StructuredResult:
        """For each feature the proportion of the range of the registered data
        that is covered by the reference dataset.

        See `SDMetrics <https://docs.sdv.dev/sdmetrics/metrics/metrics-glossary/rangecoverage>`_
        for further details.

        Returns:
            StructuredResult: The range of the registered dataset compared
                to the reference dataset.

        """
        results: Dict[str, Dict[str, float]] = self._calc(
            RangeCoverage.compute, wrapper=Wrapper.TYPE_SDMETRIC
        )
        structured_result = StructuredResult("range_coverage", results)
        return structured_result



"""Monitor class for interacting with data and scoring drift."""
from __future__ import annotations

import os
from pathlib import Path
from typing import Optional

from learning_machines_drift.backends import Backend, FileBackend
from learning_machines_drift.exceptions import ReferenceDatasetMissing
from learning_machines_drift.filter import Filter
from learning_machines_drift.metrics import Metrics
from learning_machines_drift.types import Dataset


class Monitor:
    """A class for monitoring data with data loading from backend and scoring
    drift scoring with metrics class.


    Attributes:
        tag (str): The tag where data for monitoring is located within backend.
        ref_dataset (Optional[Dataset]): The reference dataset.
        registered_dataset (Optional[Dataset]): The logged, registered dataset
            for drift comparison to reference dataset.

    """

    def __init__(
        self,
        tag: str,
        backend: Optional[Backend] = None,
    ) -> None:
        """Initialize monitor class.

        Args:
            tag (str): tag where data is located within backend.
            backend (Optional[Backend]): optional backend where data is stored.

        """

        if backend:
            self.backend: Backend = backend
        else:
            self.backend = FileBackend(Path(os.getcwd()).joinpath("lm-drift-data"))

        self.tag = tag
        self.ref_dataset: Optional[Dataset] = None
        self.registered_dataset: Optional[Dataset] = None

    def load_data(self, drift_filter: Optional[Filter] = None) -> Monitor:
        """Load data from backend into monitor.

        Args:
            drift_filter (Filter, optional): An optional filter with conditions
                applied to both reference and registered loaded data.

        Returns:
            Monitor: The calling `Monitor` instance with (optionally) filtered
                datasets loaded.

        """

        self.ref_dataset = self.backend.load_reference_dataset(self.tag)
        self.registered_dataset = self.backend.load_logged_dataset(self.tag)

        # Apply filter if passed
        if drift_filter is not None:
            self.ref_dataset = drift_filter.transform(self.ref_dataset)
            self.registered_dataset = drift_filter.transform(self.registered_dataset)

        return self

    @property
    def metrics(self) -> Metrics:
        """Drift metrics.

        Raises:
            ReferenceDatasetMissing: The reference dataset is `None`.
            ValueError: There is no additional registered data.

        """

        if self.ref_dataset is None:
            raise ReferenceDatasetMissing

        if self.registered_dataset is None:
            raise ValueError("A reference dataset is registered but not a new datasets")

        return Metrics(self.ref_dataset, self.registered_dataset)



"""Tests for backend."""
# pylint: disable=protected-access
import os
import pathlib
from uuid import uuid4

import pandas as pd
from pandas.testing import assert_frame_equal

from learning_machines_drift import Dataset, FileBackend
from learning_machines_drift.backends import get_identifier
from learning_machines_drift.datasets import example_dataset


def test_file_backend_reference(tmp_path: pathlib.Path) -> None:
    """Tests whether reference dataset upon saving and loading is equal."""

    # Given
    tag = "test_tag"
    features_df, labels_df, latents_df = example_dataset(100)
    reference_dataset = Dataset(features_df, labels_df, latents_df)

    # When
    backend = FileBackend(tmp_path)
    backend.save_reference_dataset(tag, reference_dataset)

    # Then
    recovered_dataset = backend.load_reference_dataset(tag)
    assert recovered_dataset.features.equals(recovered_dataset.features)
    assert recovered_dataset.labels.equals(recovered_dataset.labels)


def test_file_backend_features(  # pylint: disable=too-many-locals
    tmp_path: pathlib.Path,
) -> None:
    """Tests whether three batches of data can be logged at different
    identifiers and reloaded."""

    # Given
    tag = "test_tag"
    features_df_1, labels_df_1, latents_df_1 = example_dataset(5)
    features_df_2, labels_df_2, latents_df_2 = example_dataset(5)
    features_df_3, labels_df_3, latents_df_3 = example_dataset(5)
    expected_df = (
        pd.concat(
            [
                pd.concat([features_df_1, labels_df_1, latents_df_1], axis=1),
                pd.concat([features_df_2, labels_df_2, latents_df_2], axis=1),
                pd.concat([features_df_3, labels_df_3, latents_df_3], axis=1),
            ]
        )
        .sort_values(by="age")
        .reset_index(drop=True)
    )

    # When we log feature and labels
    backend = FileBackend(tmp_path)

    identifier_1 = uuid4()
    backend.save_logged_features(tag, identifier_1, features_df_1)
    backend.save_logged_labels(tag, identifier_1, labels_df_1)
    backend.save_logged_latents(tag, identifier_1, latents_df_1)

    identifier_2 = uuid4()
    backend.save_logged_features(tag, identifier_2, features_df_2)
    backend.save_logged_labels(tag, identifier_2, labels_df_2)
    backend.save_logged_latents(tag, identifier_2, latents_df_2)

    identifier_3 = uuid4()
    backend.save_logged_features(tag, identifier_3, features_df_3)
    backend.save_logged_labels(tag, identifier_3, labels_df_3)
    backend.save_logged_latents(tag, identifier_3, latents_df_3)

    # Then we should be able to load them all back
    recovered_dataset: Dataset = backend.load_logged_dataset(tag)
    recovered_df = recovered_dataset.unify()

    assert_frame_equal(
        expected_df,
        recovered_df.sort_values(by="age").reset_index(drop=True),
        check_exact=False,
    )


def test_get_identifier() -> None:
    """Tests whether identifier can be extracted from a file name string."""

    expected_uuid = uuid4()

    string_with_id = f"{expected_uuid}_any_other_text.csv"
    assert get_identifier(string_with_id) == expected_uuid


# def test_file_backend_labels(tmp_path: pathlib.Path) -> None:

#     # Given
#     tag = "test_tag"
#     _, labels_df_1 = generate_features_labels_latents(5)
#     _, labels_df_2 = generate_features_labels_latents(5)
#     _, labels_df_3 = generate_features_labels_latents(5)
#     _, labels_df_4 = generate_features_labels_latents(5)

#     expected_df = (
#         pd.concat(
#             [
#                 labels_df_1,
#                 labels_df_2,
#                 labels_df_3,
#                 labels_df_4,
#             ]
#         )
#         .sort_values(by="age")
#         .reset_index(drop=True)
#     )

#     # When
#     backend = FileBackend(tmp_path)
#     backend.save_logged_labels(tag, labels_df_1)
#     backend.save_logged_labels(tag, labels_df_2)
#     backend.save_logged_labels(tag, labels_df_3)
#     backend.save_logged_labels(tag, labels_df_4)

#     # Then
#     recovered_df = backend.load_logged_labels(tag)
#     assert_frame_equal(
#         expected_df,
#         recovered_df.sort_values(by="age").reset_index(drop=True),
#         check_exact=False,
#     )


def test_clear_logged_dataset(tmp_path: pathlib.Path) -> None:
    """Tests whether logged dataset is cleared."""
    # Given
    tag = "test_tag"

    # When
    backend = FileBackend(tmp_path)
    features_df_1, labels_df_1, latents_df_1 = example_dataset(5)
    identifier_1 = uuid4()
    backend.save_logged_features(tag, identifier_1, features_df_1)
    backend.save_logged_labels(tag, identifier_1, labels_df_1)
    backend.save_logged_latents(tag, identifier_1, latents_df_1)

    logged_path = backend._get_logged_path(tag)
    assert len(os.listdir(logged_path)) > 0
    assert backend.clear_logged_dataset(tag)
    assert len(os.listdir(logged_path)) == 0


def test_clear_reference_dataset(tmp_path: pathlib.Path) -> None:
    """Tests whether a reference dataset is cleared."""
    # Given
    tag = "test_tag"
    features_df, labels_df, latents_df = example_dataset(100)
    reference_dataset = Dataset(features_df, labels_df, latents_df)

    # When
    backend = FileBackend(tmp_path)
    backend.save_reference_dataset(tag, reference_dataset)

    reference_path = backend._get_reference_path(tag)
    assert len(os.listdir(reference_path)) > 0
    assert backend.clear_reference_dataset(tag)
    assert len(os.listdir(reference_path)) == 0



"""Class for scoring drift between reference and registered datasets."""

from typing import Any, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from learning_machines_drift.types import StructuredResult


class Display:
    """A class for converting a dictionary of drift scores to displayed output."""

    @classmethod
    def plot(
        cls,
        result: StructuredResult,
        score_name: Optional[str] = None,
        score_type: str = "pvalue",
        alpha: float = 0.05,
    ) -> Tuple[plt.Figure, Any]:
        """Plot method for displaying a set of scores on a subplot grid.

        Args:
            result (StructuredResult): Structured result from a drift
                score measurement.
            score_type (str): Either "statistic" or "pvalue".
            score_name (str): Name of score to be plotted and used as plot title.
            alpha (float): Value of alpha to be used in p-value plots.

        Returns:
            Tuple[plt.Figure, Any]: tuple of fig and subplot array.
        """
        # Set-up plots
        fig: plt.Figure[...] = plt.figure(figsize=(5, 4))
        axs: plt.Axes[...] = fig.subplots(1, 1, squeeze=False)

        # Get lists of values to use in plots
        x_vals, y_vals, xticklabels, colors = [], [], [], []
        for i, (key, scores) in enumerate(result.results.items()):
            try:
                xticklabels.append(key)
                x_vals.append(i)
                y_vals.append(scores[score_type])
                colors.append(f"C{0}")
            except KeyError:
                print(f"'{score_type}' not in result.")
                y_vals.append(np.nan)

        # Plot
        ax = axs[0, 0]
        ax.scatter(x_vals, y_vals, marker="o", label="_no_label_", color=colors)

        if score_type == "pvalue":
            ax.hlines(
                min(x_vals),
                max(x_vals),
                alpha,
                ls=":",
                label=r"$\alpha$" f"={alpha:.3f}",
            )

        # Labels
        ax.set(
            xlabel="Variable",
            ylabel=score_type,
            title=score_name,
            xticks=x_vals,
        )
        ax.set_xticklabels(xticklabels, rotation=45, ha="right")

        if score_type == "pvalue":
            ax.legend(prop={"size": "small"})

        # Return figure and axes
        return fig, axs

    @classmethod
    def table(cls, result: StructuredResult, verbose: bool = True) -> pd.DataFrame:
        """Gets a pandas dataframe and optionally prints a table of results from
        drift scoring.

        Args:
            structured_result (StructuredResult): Structured result from a drift
                score measurement.

        Returns:
            pd.DataFrame: Dataframe of scores.
        """
        # Convert dict to pandas dataframe
        df: pd.DataFrame = pd.DataFrame.from_dict(result.results, orient="index")

        # Print to stdout if verbose
        if verbose:
            print(df.to_markdown())

        # Return the dataframe version
        return df



"""Demo datasets for developing package"""

from learning_machines_drift import datasets


def test_logistic_model() -> None:
    """Test a simple generative classification model"""

    # Given the size sample we want (n)
    n_count = 10

    # When we draw from a model
    x_coord, y_coord, latents_coord = datasets.logistic_model(size=n_count)

    # Then we get n features and n labels back
    assert x_coord.shape[0] == 10
    assert y_coord.shape[0] == 10
    if latents_coord is not None:
        assert latents_coord.shape[0] == 10



"""Module for registry handling storage and logging of datasets."""
from __future__ import annotations

import os
from pathlib import Path
from typing import Optional
from uuid import UUID, uuid4

import pandas as pd

from learning_machines_drift.backends import Backend, FileBackend
from learning_machines_drift.exceptions import ReferenceDatasetMissing
from learning_machines_drift.types import (
    BaselineSummary,
    Dataset,
    FeatureSummary,
    LabelSummary,
    LatentSummary,
    ShapeSummary,
)


class Registry:
    """Class for registry for logging datasets.

    Attributes:
        backend (Optional[Backend]): Optional backend for data.
        tag (str): Tag identifying dataset.
        ref_dataset (Optional[Dataset]): Optional reference dataset.
        registered_features (Optional[pd.DataFrame]): Optional registered
            features.
        registered_labels (Optional[pd.Series]): Optional registered labels.
        registered_latents (Optional[pd.Series]): Optional registered latents.
        expect_features (bool): Whether features are expected in registry.
        expect_labels (bool): Whether a labels series is expected in registry.
        expect_latent (bool): Whether latents are expected in registry.

    """

    def __init__(
        self,
        tag: str,
        expect_features: bool = True,
        expect_labels: bool = True,
        expect_latent: bool = False,
        backend: Optional[Backend] = None,
        clear_logged: bool = False,
        clear_reference: bool = False,
    ):
        """Initializes a registry.


        Args:
            tag (str): A tag to be used with in backend for the registry.
            expect_features (bool): Should features be present.
            expect_labels (bool): Should labels be present.
            expect_latent (bool): Should latents be present.
            backend (Optional[Backend]): An optional backend to be used for
                storage.
            clear_logged (bool): Whether any existing registered data should at
                `tag` in `backend` should be cleared.
            clear_reference (bool): Whether any existing reference data at
                `tag` in `backend` should be cleared.

        """

        if backend:
            self.backend: Backend = backend
        else:
            self.backend = FileBackend(Path(os.getcwd()).joinpath("lm-drift-data"))

        self.tag: str = tag

        if clear_logged:
            self.backend.clear_logged_dataset(self.tag)
        if clear_reference:
            self.backend.clear_reference_dataset(self.tag)

        # A unique identifier used to match logged features and labels
        self._identifier: Optional[UUID] = None
        self.ref_dataset: Optional[Dataset] = None

        self.expect_features: bool = expect_features
        self.expect_labels: bool = expect_labels
        self.expect_latent: bool = expect_latent

        # Registrations
        self.registered_features: Optional[pd.DataFrame] = None
        self.registered_labels: Optional[pd.Series] = None
        self.registered_latent: Optional[pd.DataFrame] = None

    @property
    def identifier(self) -> UUID:
        """Gets the identifier of the registry.

        Returns:
            UUID: The identifier.

        """

        if self._identifier is None:
            raise ValueError("DriftDetector must be used in a context manager")

        return self._identifier

    def register_ref_dataset(
        self,
        features: pd.DataFrame,
        labels: pd.Series,
        latents: Optional[pd.DataFrame] = None,
    ) -> None:
        """Registers passed reference data.

        Args:
            features (pd.DataFrame): Reference features to be stored.
            labels (pd.Series): Reference labels to be stored.
            latents (Optional[pd.DataFrame]): Reference latents to be stored.

        """

        self.ref_dataset = Dataset(features=features, labels=labels, latents=latents)

        self.backend.save_reference_dataset(self.tag, self.ref_dataset)

    # TODO: add test # pylint: disable=fixme
    def save_reference_dataset(self, dataset: Dataset) -> None:
        """Registers passed reference data.

        Args:
            dataset (Dataset): Reference dataset to be stored.
        """

        self.ref_dataset = dataset

        self.backend.save_reference_dataset(self.tag, self.ref_dataset)

    def ref_summary(self) -> BaselineSummary:
        """Return a JSON describing shape of dataset feature, labels and
            latents.

        Returns:
            BaselineSummary: Summary of the dataset shapes.

        """

        if self.ref_dataset is None:
            raise ReferenceDatasetMissing

        feature_n_rows = self.ref_dataset.features.shape[0]
        feature_n_features = self.ref_dataset.features.shape[1]
        label_n_row = self.ref_dataset.labels.shape[0]
        if self.ref_dataset.latents is not None:
            latent_n_row = self.ref_dataset.latents.shape[0]
            latent_n_latents = self.ref_dataset.latents.shape[1]
            latents = LatentSummary(n_rows=latent_n_row, n_latents=latent_n_latents)
        else:
            latents = None

        return BaselineSummary(
            shapes=ShapeSummary(
                features=FeatureSummary(
                    n_rows=feature_n_rows, n_features=feature_n_features
                ),
                labels=LabelSummary(n_rows=label_n_row, n_labels=2),
                latents=latents,
            )
        )

    def log_features(self, features: pd.DataFrame) -> None:
        """Logs dataset features in registered data.

        Args:
            features (pd.DataFrame): Features dataframe to be registered.

        """

        self.registered_features = features
        self.backend.save_logged_features(
            self.tag, self.identifier, self.registered_features
        )

    # TODO: add test # pylint: disable=fixme
    def log_dataset(self, dataset: Dataset) -> None:
        """Logs dataset features in registered data.

        Args:
            dataset (Dataset): New dataset to be logged.

        """
        self.log_features(dataset.features)
        self.log_labels(dataset.labels)
        self.log_latents(dataset.latents)

    def log_labels(self, labels: pd.Series) -> None:
        """Logs dataset labels in registered data.

        Args:
            labels (pd.Series): Labels series to be registered.

        """

        self.registered_labels = labels
        self.backend.save_logged_labels(
            self.tag, self.identifier, self.registered_labels
        )

    def log_latents(self, latent: pd.DataFrame) -> None:
        """Logs dataset latents in registered data.

        Args:
            latents (pd.DataFrame): Latents dataframe to be registered.

        """

        self.registered_latent = latent
        self.backend.save_logged_latents(
            self.tag, self.identifier, self.registered_latent
        )

    def all_registered(self) -> bool:
        """Checks whether all expected datastes are registered.

        Returns:
            bool: True if all expected registered, False otherwise.

        """

        if self.expect_features and self.registered_features is None:
            return False

        if self.expect_labels and self.registered_labels is None:
            return False

        if self.expect_latent and self.registered_latent is None:
            return False

        return True

    @property
    def registered_dataset(self) -> Dataset:
        """Gets the registered dataset.

        Returns:
            Dataset: The registered dataset.

        """

        # This should check these two things are not None

        return Dataset(
            self.registered_features, self.registered_labels, self.registered_latent
        )

    def __enter__(self) -> Registry:
        """Assigns a new UUID upon context.

        Returns:
            Registry: With new UUID assigned to `_identifier`.

        """

        self._identifier = uuid4()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:  # type: ignore
        """Sets `_identifier` as `None` upon leaving context.

        Returns:
            Registry: With new UUID assigned to `_identifier`.

        """
        self._identifier = None



"""Tests for overall drift package."""
from learning_machines_drift import __version__


def test_version() -> None:
    """Tests whether version is expected value."""
    assert __version__ == "0.1.0"



"""Datasets module with functions for generating example data.
"""

# pylint: disable=invalid-name

from typing import Optional, Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy import stats
from scipy.special import expit


def logistic_model(
    x_mu: NDArray[np.float64] = np.array([0.0, 0.0, 0.0]),
    x_scale: NDArray[np.float64] = np.array([1.0, 1.0, 1.0]),
    x_corr: NDArray[np.float64] = np.array(
        [[1.0, 0.4, 0.0], [0.4, 1.0, 0.0], [0.0, 0.0, 1.0]]
    ),
    alpha: float = 0.5,
    beta: NDArray[np.float64] = np.array(
        [
            1.0,
            0.5,
            0.0,
        ]
    ),
    size: int = 50,
    seed: Optional[int] = None,
    return_latents: bool = False,
) -> Tuple[NDArray[np.float64], NDArray[np.float64], Optional[NDArray[np.float64]]]:
    """Generate synthetic features, labels and latents.

    Features are generated from a multivariate normal distribution, where the
    mean vector, scale vector and correlation matrix can be specified, allowing
    users to simulate covariate drift.

    Labels are generated with a logistic regression model. The regression
    parameters are controlled with the `beta` parameter, allowing simulation of
    concept drift.

    Latents are a single feature as characterizing the Bernoulli probability
    generated by the model.

    Args:
        x_mu (NDArray[np.float64]): Mean vector of features.
            Defaults to `np.array([0.0, 0.0, 0.0])`.
        x_scale (NDArray[np.float64]): Scale of features.
            Defaults to `np.array([1.0, 1.0, 1.0])`.
        x_corr (NDArray[np.float64]): Correlation matrix giving the correlation
            between features. Defaults to
            `np.array([[1.0, 0.4, 0.0], [0.4, 1.0, 0.0], [0.0, 0.0, 1.0]])`.
        alpha (float): Regression alpha parameter. Defaults to 0.5.
        beta (NDArray[np.float64]): Regression beta parameters . Defaults to
            `np.array([1.0, 0.5, 0.0])`.
        size (int): Number of samples to draw from model. Defaults to 50.
        return_latents (bool): Return underlying prediction value before
            thresholding as 'latent' data. Defaults to False.


    Returns:
        Tuple[NDArray[np.float64], NDArray[np.float64], Optional[NDArray[np.float64]]]:
            Tuple of features, labels and (optional) latents generated.
    """

    # pylint: disable=too-many-instance-attributes

    if seed is not None:
        np.random.seed(seed)

    # Sample features from variate normal
    X_diag: NDArray[np.float64] = np.diag(x_scale)
    X_covariance = np.dot(np.dot(X_diag, x_corr), X_diag)
    X = stats.multivariate_normal.rvs(x_mu, X_covariance, size=size)

    # Regression parameters
    alpha = 0.5
    beta = np.array(
        [
            1.0,
            0.5,
            0.0,
        ]
    )

    # Sample from bernoulli distribution
    theta = expit(alpha + np.dot(X, beta))
    Y = stats.bernoulli.rvs(theta)

    # If not return_latents
    if not return_latents:
        return (X, Y, None)

    return (X, Y, theta)


def example_dataset(
    n_rows: int, seed: Optional[int] = None
) -> Tuple[pd.DataFrame, pd.Series, pd.DataFrame]:

    """Generates data and returns features, labels and latents.

    Args:
        n_rows (int): Number of rows/samples.
        seed (Optional[int]): Random seed for reproducibly generating data.

    Returns:
        Tuple[pd.DataFrame, pd.Series, pd.DataFrame]: A dataset tuple of
            generated features, labels and latents.

    """

    features, labels, latents = logistic_model(
        x_mu=np.array([0.0, 0.0, 0.0]), size=n_rows, return_latents=True, seed=seed
    )

    features_df: pd.DataFrame = pd.DataFrame(
        {"age": features[:, 0], "height": features[:, 1], "ground-truth-label": labels}
    )

    predictions_series: pd.Series = pd.Series(labels, name="predicted-label")
    latents_df: pd.DataFrame = pd.DataFrame({"latents": latents})
    return (features_df, predictions_series, latents_df)



"""Tools for measuring data drift."""

from learning_machines_drift.backends import FileBackend
from learning_machines_drift.display import Display
from learning_machines_drift.filter import Filter
from learning_machines_drift.monitor import Monitor
from learning_machines_drift.registry import Registry
from learning_machines_drift.types import Dataset, StructuredResult

__version__ = "0.1.0"
__all__ = [
    "FileBackend",
    "Registry",
    "Monitor",
    "Dataset",
    "Display",
    "Filter",
    "StructuredResult",
]



"""Backend module.
"""
import glob
import os
import re
from pathlib import Path
from typing import Dict, List, Optional, Protocol, Union
from uuid import UUID

import pandas as pd

from learning_machines_drift.types import Dataset

UUIDHex4 = re.compile(
    "^([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})", re.I
)

RE_LABEL = re.compile("(labels)", re.I)
RE_FEATURES = re.compile("(features)", re.I)
RE_LATENTS = re.compile("(latents)", re.I)

DEFAULT_LABELS_NAME = "predicted-labels"


def get_identifier(path_object: Union[str, Path]) -> Optional[UUID]:
    """Extract the UUID from the filename. The filename should have the
    format UUID + some other text and a file extension. The UUID should match
    the regex in the pattern variable `UUIDHex4`.

    Args:
        path_obejct (Union[str, Path]):

    Returns:
        Optional[UUID]: Optional universally unique identifier (UUID) from
            `path_object`.

    """

    a_match = UUIDHex4.match(Path(path_object).stem)

    if a_match is None:
        return None

    return UUID(a_match.groups()[0])


class Backend(Protocol):
    """A protocol class for a Backend."""

    def save_reference_dataset(self, tag: str, dataset: Dataset) -> None:
        """Saves passed `dataset` to backend under `tag`.

        Args:
            tag (str): A tag for locating the dataset within the backend.
            dataset (Dataset): Reference dataset to be saved.

        """

    def load_reference_dataset(self, tag: str) -> Dataset:
        """Load reference dataset from reference path.

        Args:
            tag (str): Tag identifying dataset.

        """

    def save_logged_features(
        self, tag: str, identifier: UUID, dataframe: pd.DataFrame
    ) -> None:
        """Save logged features using tag as the path with UUID prepended to
        filename.

        Args:
            tag (str): Tag identifying dataset.
            identifier (UUID): A unique identifier for the logged dataset.
            dataframe (pd.DataFrame): The dataframe that needs saving.

        """

    def save_logged_labels(self, tag: str, identifier: UUID, labels: pd.Series) -> None:
        """Save logged labels using tag as the path with UUID prepended to
        filename.

        Args:
            tag (str): Tag identifying dataset.
            identifier (UUID): A unique identifier for the labels of the
                dataset.
            labels (pd.Series): The dataframe that needs saving.

        """

    def save_logged_latents(
        self, tag: str, identifier: UUID, dataframe: pd.DataFrame
    ) -> None:
        """Save optionally passed latents `dataframe` using tag as the path
        with UUID prepended to filename.

        Args:
            tag (str): Tag identifying dataset.
            identifier (UUID): A unique identifier for the labels of the
                dataset.
            dataframe (pd.DataFrame): The dataframe of latents to be saved.

        """

    def load_logged_dataset(self, tag: str) -> Dataset:
        """Return a Dataset from the union of logged data.

        Args:
            tag (str): Tag identifying dataset.

        """

    def clear_reference_dataset(self, tag: str) -> bool:
        """Delete directory containing reference files.

        Args:
            tag (str): Path to reference directory.

         Return:
            True: if `tag/reference` path exists.
            False: if `tag/reference` path does not exist.
        """

    def clear_logged_dataset(self, tag: str) -> bool:
        """Delete directory containing logged files.

        Args:
            tag (str): Path to logged directory.

        Return:
            True: if `tag/logged` path exists.
            False: if `tag/logged` path does not exist.
        """


class FileBackend:
    """Implements the Backend protocol for writing files to the filesystem."""

    def __init__(self, root_dir: Union[str, Path]) -> None:
        """Creates root directory for output
        Args:
            root_dir (Union[str, Path]): Absolute path to where outputs will be
                saved.

        Returns:
            None
        """
        self.root_dir = Path(root_dir)
        self.root_dir.mkdir(exist_ok=True)

    def _get_reference_path(self, tag: str) -> Path:
        """Get the directory for storing reference data, creating directories
        if required.

        Args:
            tag (str): Tag identifying dataset

        Returns:
            Path: Directory storing reference datasets
        """

        # Make a directory for the current tag
        tag_dir = self.root_dir.joinpath(tag)
        if not tag_dir.exists():

            tag_dir.mkdir()

        # Make a directory for the reference data
        reference_dir = tag_dir.joinpath("reference")
        if not reference_dir.exists():
            reference_dir.mkdir()

        return reference_dir

    def _get_logged_path(self, tag: str) -> Path:
        """Get the directory for storing logged data, creating directories if
        required.

        Args:
            tag (str): Tag identifying dataset.

        Returns:
            Path: Directory storing logged datasets.
        """

        # Make a directory for the current tag
        tag_dir = self.root_dir.joinpath(tag)
        if not tag_dir.exists():
            tag_dir.mkdir()

        # Make a directory for the reference data
        logged_dir = tag_dir.joinpath("logged")
        if not logged_dir.exists():
            logged_dir.mkdir()

        return logged_dir

    def save_reference_dataset(self, tag: str, dataset: Dataset) -> None:
        """Saves passed `dataset` to backend under `tag`.

        Args:
            tag (str): A tag for locating the dataset within the backend.
            dataset (Dataset): Reference dataset to be saved.

        """
        reference_dir = self._get_reference_path(tag)
        dataset.features.to_csv(reference_dir.joinpath("features.csv"), index=False)
        if dataset.labels.name is None:
            dataset.labels.name = DEFAULT_LABELS_NAME
        labels_dataframe = pd.DataFrame(dataset.labels)
        labels_dataframe.to_csv(reference_dir.joinpath("labels.csv"), index=False)
        if dataset.latents is not None:
            dataset.latents.to_csv(reference_dir.joinpath("latents.csv"), index=False)

    def load_reference_dataset(self, tag: str) -> Dataset:
        """Load reference dataset from reference path.

        Args:
            tag (str): Tag identifying dataset.

        """
        reference_dir = self._get_reference_path(tag)

        features_df: pd.DataFrame = pd.read_csv(reference_dir.joinpath("features.csv"))
        labels_df: pd.DataFrame = pd.read_csv(reference_dir.joinpath("labels.csv"))
        labels_series: pd.Series = labels_df.iloc[:, 0]
        latents_df: Optional[pd.DataFrame] = None
        if reference_dir.joinpath("latents.csv").exists():
            latents_df = pd.read_csv(reference_dir.joinpath("latents.csv"))

        return Dataset(features_df, labels_series, latents_df)

    def save_logged_features(
        self, tag: str, identifier: UUID, dataframe: pd.DataFrame
    ) -> None:
        """Save logged features using tag as the path with UUID prepended to
        filename.

        Args:
            tag (str): Tag identifying dataset.
            identifier (UUID): A unique identifier for the logged dataset.
            dataframe (pd.DataFrame): The dataframe that needs saving.

        """
        logged_dir = self._get_logged_path(tag)
        dataframe.to_csv(logged_dir.joinpath(f"{identifier}_features.csv"), index=False)

    def save_logged_labels(self, tag: str, identifier: UUID, labels: pd.Series) -> None:
        """Save logged labels using tag as the path with UUID prepended to
        filename.

        Args:
            tag (str): Tag identifying dataset.
            identifier (UUID): A unique identifier for the labels of the
                dataset.
            labels (pd.Series): The dataframe that needs saving.

        """
        logged_dir = self._get_logged_path(tag)
        if labels.name is None:
            labels.name = DEFAULT_LABELS_NAME
        labels_dataframe = pd.DataFrame(labels)

        labels_dataframe.to_csv(
            logged_dir.joinpath(f"{identifier}_labels.csv"), index=False
        )

    def save_logged_latents(
        self, tag: str, identifier: UUID, dataframe: Optional[pd.DataFrame]
    ) -> None:
        """Save optionally passed latents `dataframe` using tag as the path
        with UUID prepended to filename.

        Args:
            tag (str): Tag identifying dataset.
            identifier (UUID): A unique identifier for the labels of the
                dataset.
            dataframe (pd.DataFrame): The dataframe of latents to be saved.

        """
        if dataframe is not None:
            logged_dir = self._get_logged_path(tag)
            dataframe.to_csv(
                logged_dir.joinpath(f"{identifier}_latents.csv"), index=False
            )

    def load_logged_dataset(self, tag: str) -> Dataset:
        """Return a Dataset from the union of logged data.

        Args:
            tag (str): Tag identifying dataset.

        """
        files = [Path(f) for f in glob.glob(f"{self._get_logged_path(tag)}/*")]
        loaded_file_dict: Dict[UUID, List[Path]] = {}

        for file in files:
            # Get its identifier
            key = get_identifier(Path(file))
            if key is None:
                raise IOError("File name does not start with a UUID")

            loaded_file_dict.setdefault(key, []).append(file)

        all_feature_dfs: List[pd.DataFrame] = []
        all_label_series: List[pd.Series] = []
        all_latent_dfs: List[pd.DataFrame] = []

        for key, value in loaded_file_dict.items():
            # Must have at least features and labels, optional latents
            assert len(value) >= 2

            # Loop over files in list of files for each identifier
            for fname in value:
                if RE_LABEL.search(fname.stem) is not None:
                    all_label_df: pd.DataFrame = pd.read_csv(fname, header=0)
                    assert len(all_label_df.columns) == 1
                    all_label_series.append(all_label_df.iloc[:, 0])

                if RE_FEATURES.search(fname.stem) is not None:
                    all_feature_dfs.append(pd.read_csv(fname))

                if RE_LATENTS.search(fname.stem) is not None:
                    all_latent_dfs.append(pd.read_csv(fname))

        # Assert that the features and labels are non-empty
        assert all_feature_dfs and all_label_series

        # If latents found, return with latents, otherwise no latents
        features: pd.DataFrame = pd.concat(all_feature_dfs).reset_index(drop=True)
        labels: pd.Series = pd.concat(all_label_series).reset_index(drop=True)
        latents: Optional[pd.DataFrame] = (
            pd.concat(all_latent_dfs).reset_index(drop=True) if all_latent_dfs else None
        )
        return Dataset(features=features, labels=labels, latents=latents)

    def clear_reference_dataset(self, tag: str) -> bool:
        """Delete directory containing reference files.

        Args:
            tag (str): Path to reference directory.

         Return:
            True: if `tag/reference` path exists.
            False: if `tag/reference` path does not exist.
        """
        reference_dir = self.root_dir.joinpath(tag).joinpath("reference")
        if not reference_dir.exists():
            return False
        if len(os.listdir(reference_dir)) == 0:
            return True

        for root, _, files in os.walk(reference_dir):
            for file in files:
                os.remove(os.path.join(root, file))
        return True

    def clear_logged_dataset(self, tag: str) -> bool:
        """Delete directory containing logged files.

        Args:
            tag (str): Path to logged directory.

        Return:
            True: if `tag/logged` path exists.
            False: if `tag/logged` path does not exist.
        """
        logged_dir = self.root_dir.joinpath(tag).joinpath("logged")
        if not logged_dir.exists():
            return False
        if len(os.listdir(logged_dir)) == 0:
            return True

        for root, _, files in os.walk(logged_dir):
            for file in files:
                os.remove(os.path.join(root, file))
        return True
